#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
//#include <math.h>
//#include <float.h>
#include <string.h>

#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>


#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <errno.h>
#include <unistd.h>
//#include <sys/time.h>



#define IsPort 			0b00000001
#define IsDirectory		0b00000010
#define IsTreshold		0b00000100
#define IsAllValuesOk	0b00000111

//Коды машины состояний
#define WaitingForWeight	1
#define StartNewLog			2
#define	Logging				3
#define StopLogging			4


#define DataReady		1
#define Timeout			0
#define PacketSize		22

//Глобальные переменные
char PORT[32]; //хранит имя заданного com-порта
static int  pcom = -1; //хранит дескриптор файла порта

float TRESHOLD;//Хранит порог веса, значения выше или равные которому будут записаны


//-----------------------------Функции--------------------------------------------//

//Открывает порт и применяет нужные настройки - биты, скорость...
void open_uart()
{
	//Структура с помощью которой производится настройка параметров UART.
	static struct termios tty; 

	 if ((pcom = open(PORT, O_RDWR | O_NDELAY | O_NOCTTY))==-1)
	{
		fprintf(stderr, "Порт %s не может быть открыт, ошибка: ", PORT);
		perror(NULL);
	}
	tcflush(pcom, TCIOFLUSH);
	tcgetattr(pcom, &tty);

	#ifndef __CYGWIN__
		cfmakeraw(&tty);
	#else
		tty.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP
		            | INLCR | IGNCR | ICRNL | IXON);
		tty.c_oflag &= ~OPOST;
		tty.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
		tty.c_cflag &= ~(CSIZE | PARENB);
		tty.c_cflag |= CS8;
	#endif
		tty.c_cflag &= ~CRTSCTS;
		tcsetattr(pcom, TCSANOW, &tty);
		tcflush(pcom, TCIOFLUSH);
		tcsetattr(pcom, TCSANOW, &tty);
		tcflush(pcom, TCIOFLUSH);
		tcflush(pcom, TCIOFLUSH);
		cfsetospeed(&tty, B9600);
		cfsetispeed(&tty, B9600);
		tcsetattr(pcom, TCSANOW, &tty);
}

void close_uart()
{
	if (pcom!=-1) close(pcom);
}

//Если в приемном буффере есть несчитанные байты, то возвражает DataReady,
//если нет, то ожидает прихода их указанное количество секунд.
//Не пришли - возвращает Timeout. Пришли - DataReady.
int WaitForByteWithTimeout(int sec)
{
    fd_set rec;
    struct timeval t;
	int n = 0;

    FD_ZERO(&rec);
    FD_SET(pcom,&rec);

    t.tv_sec  = sec;
    t.tv_usec = 0;

    n = select(pcom+1,&rec,NULL,NULL,&t);
	if (n<0)
		{
		fprintf(stderr,"Системный вызов select завершился с ошибкой !\n");
		return Timeout;
		}
	else if (n==0)
		return Timeout;
	else
		{
		if (FD_ISSET(pcom,&rec)) return DataReady;
		}
    return Timeout;
}
//Ждет прихода пакета 3 секунды. В случае прихода пакета возвращает управление сразу.
//Возвращаемое значение =DataReady, а по указателю outputbuffer записывается принятый
//пакет. Если пакета нет 3 секунды, то возвращает Timeout.
int WaitForPacketWithTimeout(unsigned char * outputbuffer)
{
	unsigned char byte;
	static unsigned char buf[256];
	static unsigned char index = 0;
	int count;
	
	while (WaitForByteWithTimeout(3))
		{
		count = read(pcom, &byte, 1);
		if (count == 1) buf[index++] = byte;
		if (index>=PacketSize)
			{
			 //53 54 2C 47 53 2C 30 BE 2C 20 20 20 20 20 20 20 30 20 6B 67 0D 0A
			 // Характерные признаки пакета:
			 // зпт-2,5,8 пробел-17 k-18 g-19 CR-20 LF-21
			 // [2,5,8]=0x2C	[17]=0x20	[18]=0x6B	[19]=0x67	[20]=0x0D	[21]=0x0A
			if ((buf[index-20] == 0x2C)&&(buf[index-17] == 0x2C)&&(buf[index-14] == 0x2C)&&\
			(buf[index-5] == 0x20)&&(buf[index-4] == 0x6B)&&(buf[index-3] == 0x67)&&\
			(buf[index-2] == 0x0D)&&(buf[index-1] == 0x0A))
				{
				// начиная с index-PacketSize в буффере есть валидный пакет
				memcpy(outputbuffer,&buf[index-PacketSize],PacketSize);
				index = 0; //Отбрасываем возможно накопленный мусор.
				return DataReady;
				}
			}
		}
	return Timeout;
}
// Проверяет посланные программе аргументы командной строки,
// наличие в системе заданного порта, пути для сохранения логов.
int parse_arguments (int argc, char ** argv)
{
    int i = 0;
	static int errsve;
	int opt;
	int fd;
	char * control_ptr;
    char help_str[] =
	"\nИспользование: weightlogger -port порт -dir путь_к_папке -trh порог\n"
	"Где:\n"
    "порт - ваш порт к которому подключено устройство, прим.: /dev/ttyS0\n"
    "путь_к_папке - путь к папке логов, прим.: /home/name/weightlogs/\n"
    "порог - вес (кг) выше или равный этому будет сохранен, прим.: 40\n";
       const struct option long_opts[] = {
              { "port", required_argument, NULL, '1' },
              { "dir", required_argument, NULL, '2' },
              { "trh", required_argument, NULL, '3' },
              { NULL, 0, NULL, 0 }
       };

       if (argc==1)
			{
			printf("%s",help_str);
			return 1;
			}
       while ((opt = getopt_long_only (argc, argv, "", long_opts, NULL)) != -1) {
              switch (opt) {
                     case '1':
						strcpy(PORT, optarg);
						 if ((fd = open(PORT, O_RDWR | O_NDELAY | O_NOCTTY))!=-1)
						{
							i|=IsPort; //Значение принято
							//printf ("Порт %s найден - OK !\n", PORT);
							close(fd);
						}
						else
						{
							errsve = errno;
							control_ptr = strerror(errsve);
							fprintf(stderr, "Порт %s не может быть открыт, ошибка:  %s\n", optarg, control_ptr);
						};
						         break;

                     case '2':
						 if (!chdir(optarg)) //Попытка смены текущего каталога на предоставленный пользователем.
						{
							i|=IsDirectory; //Успешно.
							//printf ("Каталог %s существует !\n", optarg);
						}
						else
						{
							errsve = errno;
							control_ptr = strerror(errsve);
							fprintf(stderr,"Путь %s не может быть открыт, ошибка: %s\n",optarg, control_ptr);
						};
						         break;

                     case '3':
						 TRESHOLD = strtof(optarg,&control_ptr);
						 if ((control_ptr!=optarg)&&(*control_ptr=='\0'))
						{
							i|=IsTreshold; //Значение принято
							//printf ("Порог=%f Kg\n", TRESHOLD);
						}
						else
						{
							fprintf(stderr,"\"%s\" не числовое значение !\n",optarg);
						};
						         break;

                     case '?':
                     fprintf (stderr, "%s\n", help_str);
                     return 1;

                     default:
                     fprintf (stderr, "Непонятная ошибка !\n");
                     return 1;
              }
       }


       if (!(i==IsAllValuesOk))
       {
	       fprintf(stderr,"\nЕсть пропущенные или некорректные аргументы, смотри помощь:\n%s",help_str);
	       return 1;
       }
       return 0;
}

//Формирует строку с текущим временем в формате ГГГГ.ММ.ДД_ЧЧ.мм.сс пример 2014.09.13_17.19.44
int GetCurrentTimeString(char * output)
{
       time_t t;
       struct tm *tmp;
       t = time(NULL);
       tmp = localtime(&t);
       if (tmp == NULL) {
           perror("localtime");
           return -1;
       }
		//Двавцать символов, включая завершающий null
       if (strftime(output, 20, "%Y.%m.%d@%H.%M.%S", tmp) == 0) {
           fprintf(stderr, "Вызов strftime завершился с ошибкой !\n");
           return -1;
       }
		return 0;
}

//Конвертирует содержащееся в посылке символьное представление веса в формат с плавающей точкой
//  !!! Возвращает абсолютное значение !!!
float GetWeightFromBuffer(char * buf)
{
	static char str[8] = {0,0,0,0,0,0,0,0}; //8-й символ должен быть 0
	char * control_ptr;
	float tempfloat;
	//В буффере, начиная с 10 (если начало считать с 0-го) байта содержится 7 символов веса.(Без "-")
	memcpy(str,&buf[10],7); 
	tempfloat=strtof(str,&control_ptr);
	     if ((control_ptr!=str)&&(*control_ptr=='\0'))
		{
			return tempfloat;
		}
		else
		{
			fprintf(stderr, "\"%s\" не числовое значение веса в пакете !\n",str);
			return 0.0f; //Если вдруг строку не удалось преобразовать, то возвратим 0
		}
}

//--------------------------Основная программа-----------------------------//
int main (int argc, char ** argv)
{
	int i = 0;
	FILE * logfile; //Указатель текущего лог файла
	static int State = WaitingForWeight; //Начальное состояние - ожидание подходящего веса.
	float Weight; //Вес
	unsigned char TresholdIntegrator = 0; //Сколько раз зарегистрирован вес выше или равный пороговому.
	unsigned char Packet[PacketSize]; //Последний принятый пакет
	unsigned char tempBuffer[PacketSize*8]; //Временный буффер под 8 последних пакетов
	int tempbufferindex = 0;
	char * control;
	char CurrTimeNDate [32];// Строка даты и времени используется GetCurrentTimeString

	if (parse_arguments(argc, argv)) return 1;
	open_uart();
	if (GetCurrentTimeString(CurrTimeNDate)) return 1;

	//Машина состояний

	while (1)
		{
		switch (State)
			{
			case WaitingForWeight:
				if (WaitForPacketWithTimeout(Packet) == DataReady)
					{
					Weight = GetWeightFromBuffer(Packet);
					memcpy(&tempBuffer[tempbufferindex*PacketSize],Packet,PacketSize);//копируем во временный буффер.
					tempbufferindex++; tempbufferindex&=7; //кольцуем индекс.
					if (Weight>=TRESHOLD)
						{
						TresholdIntegrator++; //Накапливаем количество пакетов подряд с весом больше порога.
						if (TresholdIntegrator>8) TresholdIntegrator = 8; //Ограничиваем максимальное значение 8-ю.
						}
					else
						{
						if (TresholdIntegrator) TresholdIntegrator--;// Если и так 0, то не отнимаем.
						}
					if (TresholdIntegrator == 8) State = StartNewLog; //Переходим к началу записи в файл.
					}
					else
					{
					//Связь оборвалась. Сбрасываем счетчик.
					TresholdIntegrator = 0;
					}
			break;

			case StartNewLog:
				GetCurrentTimeString(CurrTimeNDate);//Получаем текущее время и дату.
				strcat(CurrTimeNDate,".txt");//Добавляем за ним .txt и 
				logfile = fopen(CurrTimeNDate,"a");//пытаемся создать файл с таким именем.
				if (logfile == NULL)
					{
					control = strerror(errno);
					fprintf(stderr, "Файл %s не может быть создан, ошибка:  %s\n",CurrTimeNDate, control);
					}
				else
					{
					for (i = 0;i<8;i++)
						{
						//Записываем в файл 8 пакетов в порядке их прихода.
						fwrite(&tempBuffer[tempbufferindex*PacketSize],PacketSize,1,logfile);
						tempbufferindex++; tempbufferindex&=7;
						}
					}
				State = Logging; //Переходим к стадии планомерной записи в порядке прихода пакетов.
			break;

			case Logging:
				if (WaitForPacketWithTimeout(Packet) == DataReady)
					{
					Weight = GetWeightFromBuffer(Packet);
					if (Weight>=TRESHOLD)
						{
						TresholdIntegrator++; //Накапливаем количество пакетов подряд с весом больше порога.
						if (TresholdIntegrator>8) TresholdIntegrator = 8; //Ограничиваем максимальное значение 8-ю.
						}
					else
						{
						if (TresholdIntegrator) TresholdIntegrator--;// Если и так 0, то не отнимаем.
						}
					if (TresholdIntegrator == 0) State = StopLogging; //Переходим к закрытию файла.
					if (logfile != NULL) fwrite(Packet,PacketSize,1,logfile); //Сохраняем принятый пакет.
					}
				else
					{
					//Оказывается нет связи (таймаут) !
					TresholdIntegrator = 0;
					State = StopLogging; //Переходим к закрытию файла.
					}
			break;

			case StopLogging:
					if (logfile != NULL) fclose(logfile);//Закрываем файл.
					logfile = NULL;
					State = WaitingForWeight;//Переходим к ожиданию следующего веса выше порога.
			break;
			}
		};
	//Эти строки никогда не выполняются.
	close_uart();
	return 0;
}









